#include "tests.h"

#include "log.h"
#include "mem.h"
#include "ssb.tests.h"

#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#if defined(_WIN32)
#define WIFEXITED(x) 1
#define WEXITSTATUS(x) (x)
#endif

static void _test_nop(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file, "print('hi');");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	(void)result;
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);
}

static void _test_child(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.onExit = function() {\n"
		"	print('child exited');\n"
		"};\n"
		"task.activate();\n"
		"File.readFile('out/child.js').then(function(data) {\n"
		"	task.execute({name: 'child.js', source: utf8Decode(data)}).then(function() {\n"
		"		print('child started');\n"
		"	});\n"
		"});");
	fclose(file);

	file = fopen("out/child.js", "w");
	fprintf(file,
		"print('I am the child process.');\n"
		"exit(0);\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	(void)result;
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/child.js");
}

static void _test_promise(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.activate();\n"
		"File.readFile('out/child.js').then(function(data) {\n"
		"	task.execute({name: 'child.js', source: utf8Decode(data)}).then(function() {\n"
		"		task.getExports().then(function(exports) {\n"
		"			return exports.add(1, 1);\n"
		"		}).then(function(sum) {\n"
		"			if (sum == 2) {\n"
		"				exit(0);\n"
		"			} else {\n"
		"				exit(1);\n"
		"			}\n"
		"		});\n"
		"	});\n"
		"});\n");
	fclose(file);

	file = fopen("out/child.js", "w");
	fprintf(file,
		"exports = {\n"
		"	add: function(left, right) {\n"
		"		return left + right;\n"
		"	}\n"
		"}\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	(void)result;
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/child.js");
}

static void _test_promise_remote_throw(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.activate();\n"
		"File.readFile('out/child.js').then(function(data) {\n"
		"	task.execute({name: 'child.js', source: utf8Decode(data)}).then(function() {\n"
		"		task.getExports().then(function(exp) {\n"
		"			return exp.add(1, 1);\n"
		"		}).then(function(sum) {\n"
		"			exit(1);\n"
		"		}).catch(function(error) {\n"
		"			print('Caught: ' + error.message);\n"
		"			if (error.stack) {\n"
		"				print('stack: ' + error.stack);\n"
		"			}\n"
		"			exit(0);\n"
		"		});\n"
		"	}).catch(function(e) {\n"
		"		print('caught', e.message);\n"
		"	});\n"
		"});\n");
	fclose(file);

	file = fopen("out/child.js", "w");
	fprintf(file,
		"exports = {\n"
		"	add: function(left, right) {\n"
		"		throw new Error('fail');\n"
		"	}\n"
		"}\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	(void)result;
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/child.js");
}

static void _test_promise_remote_reject(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.activate();\n"
		"File.readFile('out/child.js').then(function(data) {\n"
		"	task.execute({name: 'child.js', source: utf8Decode(data)}).then(function() {\n"
		"		task.getExports().then(function(exp) {\n"
		"			return exp.add(1, 1);\n"
		"		}).then(function(sum) {\n"
		"			exit(1);\n"
		"		}).catch(function(error) {\n"
		"			print('Caught: ' + error.message);\n"
		"			if (error.stack) {\n"
		"				print('stack: ' + error.stack);\n"
		"			}\n"
		"			exit(0);\n"
		"		});\n"
		"	}).catch(function(e) {\n"
		"		print('caught', e.message);\n"
		"	});\n"
		"});\n");
	fclose(file);

	file = fopen("out/child.js", "w");
	fprintf(file,
		"exports = {\n"
		"	add: function(left, right) {\n"
		"		return new Promise(function(resolve, reject) {\n"
		"			reject(new Error('oops'));\n"
		"		});\n"
		"	}\n"
		"}\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	(void)result;
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/child.js");
}

static void _test_database(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var db = new Database('testdb');\n"
		"if (db.get('a')) {\n"
		"	exit(1);\n"
		"}\n"
		"db.set('a', 1);\n"
		"if (db.get('a') != 1) {\n"
		"	exit(2);\n"
		"}\n"
		"db.set('b', 2);\n"
		"db.set('c', 3);\n"
		"\n"
		"var expected = ['a', 'b', 'c'];\n"
		"var have = db.getAll();\n"
		"for (var i = 0; i < have.length; i++) {\n"
		"	var item = have[i];\n"
		"	if (expected.indexOf(item) == -1) {\n"
		"		print('Did not find ' + item + ' in db.');\n"
		"		exit(3);\n"
		"	} else {\n"
		"		expected.splice(expected.indexOf(item), 1);\n"
		"	}\n"
		"}\n"
		"if (expected.length) {\n"
		"	print('Expected but did not find: ' + JSON.stringify(expected));\n"
		"	exit(4);\n"
		"}\n");
	fclose(file);

	char command[256];
	unlink("out/test_db0.sqlite");
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=out/test_db0.sqlite -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/test_db0.sqlite");
}

static void _test_this(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.activate.bind(null).apply();\n"
		"exit(0);\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _test_await(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"print('hi');\n"

		"function foobar() {\n"
		"	return new Promise(function(resolve, reject) {\n"
		"		resolve(10);\n"
		"	});\n"
		"}\n"
		"\n"
		"async function huh() {\n"
		"	let v = await foobar();\n"
		"	print('v => ' + v);\n"
		"	if (v != 10) {\n"
		"		throw new Error('nope');\n"
		"	}\n"
		"}\n"
		"\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _test_import(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"import * as req from './required.js';\n"
		"if (req.foo() != 12345) {\n"
		"	exit(1);\n"
		"}\n");
	fclose(file);

	file = fopen("out/required.js", "w");
	fprintf(file,
		"export function foo() {\n"
		"	return 12345;\n"
		"}\n");
	fclose(file);

	file = fopen("out/bad.js", "w");
	fprintf(file,
		"import * as req from './missing.js';\n"
		"if (req.foo() != 12345) {\n"
		"	exit(1);\n"
		"}\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/bad.js", options->exe_path);
	tf_printf("%s\n", command);
	result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/required.js");
	unlink("out/missing.js");
}

static void _test_exit(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file, "import * as blah from './blah.js';\n");
	fclose(file);

	file = fopen("out/blah.js", "w");
	fprintf(file, "\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/blah.js");
}

static void _test_icu(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"print('Hi');\n"
		"print(parseInt('3').toLocaleString());\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _test_uint8array(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.onExit = function() {\n"
		"	print('child exited');\n"
		"};\n"
		"task.activate();\n"
		"File.readFile('out/child.js').then(function(data) {\n"
		"	task.execute({name: 'child.js', source: utf8Decode(data)}).then(async function() {\n"
		"		print('child started');\n"
		"		var input = new Uint8Array(10);\n"
		"		for (var i = 0; i < 10; i++) {\n"
		"			input[i] = i;\n"
		"		}\n"
		"		var test = (await task.getExports()).test;\n"
		"		var output = new Uint8Array(await test(input));\n"
		"		print('input', input, input.length, input.byteLength);\n"
		"		print('output', output, output.length, output.byteLength);\n"
		"		for (var i = 0; i < 10; i++) {\n"
		"			print(output[i]);\n"
		"			if (output[i] != i) {\n"
		"				print('output[' + i + '] == ' + output[i]);\n"
		"				exit(1);\n"
		"			}\n"
		"		}\n"
		"		exit(0);\n"
		"	})\n"
		"})\n");
	fclose(file);

	file = fopen("out/child.js", "w");
	fprintf(file,
		"exports = {\n"
		"	test: function(data) {\n"
		"		return data;\n"
		"	}\n"
		"}\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/child.js");
}

static void _test_float(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"var task = new Task();\n"
		"task.onExit = function() {\n"
		"	print('child exited');\n"
		"};\n"
		"task.activate();\n"
		"File.readFile('out/child.js').then(function(data) {\n"
		"	task.execute({name: 'child.js', source: utf8Decode(data)}).then(async function() {\n"
		"		print('get exports');\n"
		"		let test = (await task.getExports()).test;\n"
		"		print('calling export');\n"
		"		let result = await test(1.2);\n"
		"		print(result);\n"
		"		exit(result == 1.2 ? 0 : 1);\n"
		"	});\n"
		"});");
	fclose(file);

	file = fopen("out/child.js", "w");
	fprintf(file,
		"print(\"child\");\n"
		"exports = {\n"
		"	test: function(value) {\n"
		"		print(value);\n"
		"		return value;\n"
		"	}\n"
		"};\n"
		"print(\"child ready\");\n"
		);
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	(void)result;
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
	unlink("out/child.js");
}

static void _test_socket(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"'use strict';\n"
		"\n"
		"var s = new Socket();\n"
		"print('connecting');\n"
		"print('before connect', s.isConnected);\n"
		"s.onError(function(e) {\n"
		"	print(e);\n"
		"});\n"
		"print('noDelay', s.noDelay);\n"
		"s.noDelay = true;\n"
		"s.connect('www.unprompted.com', 80).then(function() {\n"
		"	print('connected', 'www.unprompted.com', 80, s.isConnected);\n"
		"	print(s.peerName);\n"
		"	s.read(function(data) {\n"
		"		print('read', data ? data.length : null);\n"
		"	});\n"
		"	s.write('GET / HTTP/1.0\\r\\n\\r\\n');\n"
		"}).then(function(e) {\n"
		"	print('closed 1');\n"
		"});\n"
		"\n"
		"var s2 = new Socket();\n"
		"print('connecting');\n"
		"print('before connect', s2.isConnected);\n"
		"s2.onError(function(e) {\n"
		"	print('error');\n"
		"	print(e);\n"
		"});\n"
		"print('noDelay', s2.noDelay);\n"
		"s2.noDelay = true;\n"
		"s2.connect('www.unprompted.com', 443).then(function() {\n"
		"	print('connected', 'www.unprompted.com', 443);\n"
		"	s2.read(function(data) {\n"
		"		print('read', data ? data.length : null);\n"
		"	});\n"
		"	return s2.startTls();\n"
		"}).then(function() {\n"
		"	print('ready');\n"
		"	print(s2.peerName);\n"
		"	s2.write('GET / HTTP/1.0\\r\\nConnection: close\\r\\n\\r\\n').then(function() {\n"
		"		s2.shutdown();\n"
		"	});\n"
		"}).catch(function(e) {\n"
		"	print('caught');\n"
		"	print(e);\n"
		"});\n"
		"var s3 = new Socket();\n"
		"print('connecting s3');\n"
		"print('before connect', s3.isConnected);\n"
		"s3.onError(function(e) {\n"
		"	print('error');\n"
		"	print(e);\n"
		"});\n"
		"print('noDelay', s3.noDelay);\n"
		"s3.noDelay = true;\n"
		"s3.connect('0.0.0.0', 443).then(function() {\n"
		"	print('connected', '0.0.0.0', 443);\n"
		"	s3.read(function(data) {\n"
		"		print('read', data ? data.length : null);\n"
		"	});\n"
		"	return s3.startTls();\n"
		"}).then(function() {\n"
		"	print('ready');\n"
		"	print(s3.peerName);\n"
		"	s3.write('GET / HTTP/1.0\\r\\nConnection: close\\r\\n\\r\\n').then(function() {\n"
		"		s3.shutdown();\n"
		"	});\n"
		"}).catch(function(e) {\n"
		"	print('caught');\n"
		"	print(e);\n"
		"});\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _test_file(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"'use strict';\n"
		"File.readFile('out/test.js').then(function(data) {\n"
		"}).catch(function(error) {\n"
		"	print('ERROR', error);\n"
		"	exit(1);\n"
		"});\n"
		"File.readFile('out/missing.txt').then(function(data) {\n"
		"	print('READ', utf8Decode(data));\n"
		"	exit(1);\n"
		"}).catch(function(error) {\n"
		"	print('expected error', error);\n"
		"});\n");
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _test_sign(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"'use strict';\n"
		"let id = ssb.createIdentity('test');\n"
		"print(id);\n"
		"let sig = ssb.hmacsha256sign('hello', 'test', id);\n"
		"print(sig);\n"
		"if (!ssb.hmacsha256verify(id, 'hello', sig)) {\n"
		"	exit(1);\n"
		"}\n"
		"if (ssb.hmacsha256verify(id, 'world', sig)) {\n"
		"	exit(1);\n"
		"}\n"
		"if (ssb.hmacsha256verify(id, 'hello1', sig)) {\n"
		"	exit(1);\n"
		"}\n"
		);
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _test_b64(const tf_test_options_t* options)
{
	FILE* file = fopen("out/test.js", "w");
	fprintf(file,
		"'use strict';\n"
		"print(base64Encode('hello'));\n"
		"let x = base64Decode(base64Encode('hello'));\n"
		"if (x !== 'hello') {\n"
		"	print(x);\n"
		"	exit(1);\n"
		"}\n"
		);
	fclose(file);

	char command[256];
	snprintf(command, sizeof(command), "%s run --ssb-port=0 --db-path=:memory: -s out/test.js", options->exe_path);
	tf_printf("%s\n", command);
	int result = system(command);
	tf_printf("returned %d\n", WEXITSTATUS(result));
	assert(WIFEXITED(result));
	assert(WEXITSTATUS(result) == 0);

	unlink("out/test.js");
}

static void _tf_test_run(const tf_test_options_t* options, const char* name, void (*test)(const tf_test_options_t* options), bool opt_in)
{
	bool specified = false;
	if (options->tests)
	{
		char* dup = tf_strdup(options->tests);
		char* state = NULL;
		const char* t = NULL;
		while ((t = strtok_r(t ? NULL : dup, ",", &state)) != NULL)
		{
			if (strcmp(t, name) == 0)
			{
				specified = true;
				break;
			}
		}
		tf_free(dup);
	}

	if ((!opt_in && !options->tests) || specified)
	{
#define GREEN "\e[1;32m"
#define MAGENTA "\e[1;35m"
#define CYAN "\e[1;36m"
#define RESET "\e[0m"
		tf_printf(CYAN "== running test " MAGENTA "%s" CYAN " ==\n" RESET, name);
		test(options);
		tf_printf("[" GREEN "pass" RESET "] %s\n", name);
#undef GREEN
#undef MAGENTA
#undef CYAN
#undef RESET
	}
}

void tf_tests(const tf_test_options_t* options)
{
	_tf_test_run(options, "ssb", tf_ssb_test_ssb, false);
	_tf_test_run(options, "ssb_id", tf_ssb_test_id_conversion, false);
	_tf_test_run(options, "ssb_following", tf_ssb_test_following, false);
	_tf_test_run(options, "nop", _test_nop, false);
	_tf_test_run(options, "child", _test_child, false);
	_tf_test_run(options, "promise", _test_promise, false);
	_tf_test_run(options, "promise_remote_throw", _test_promise_remote_throw, false);
	_tf_test_run(options, "promise_remote_reject", _test_promise_remote_reject, false);
	_tf_test_run(options, "database", _test_database, false);
	_tf_test_run(options, "this", _test_this, false);
	_tf_test_run(options, "await", _test_await, false);
	_tf_test_run(options, "import", _test_import, false);
	_tf_test_run(options, "exit", _test_exit, false);
	_tf_test_run(options, "icu", _test_icu, false);
	_tf_test_run(options, "uint8array", _test_uint8array, false);
	_tf_test_run(options, "float", _test_float, false);
	_tf_test_run(options, "socket", _test_socket, false);
	_tf_test_run(options, "file", _test_file, false);
	_tf_test_run(options, "sign", _test_sign, false);
	_tf_test_run(options, "b64", _test_b64, false);
	_tf_test_run(options, "rooms", tf_ssb_test_rooms, false);
	_tf_test_run(options, "bench", tf_ssb_test_bench, false);
	_tf_test_run(options, "go-ssb-room", tf_ssb_test_go_ssb_room, true);
	tf_printf("Tests completed.\n");
}
